/** Licensed Materials - Property of IBM, 5724-U69, (C) Copyright IBM Corp. 2008, 2010 - All Rights reserved.  **/
dojo.provide("com.ibm.mm.enabler.extreg");

if (typeof(ExtensionRegistry) == "undefined") {

    /**
     * Constructs a new extension registry object.
     *
     * The optional mappingFn parameter provides an opportunity to adapt the extension registry to
     * map property names on the extension definition variable (what is typically the name of the function
     * being extended) to a different function. This mapping allows the extension registry to support
     * the following use cases:
     *
     *             1. Use a public API function in Enabler as the "extension point" specified by the
     *                customer/consumer, but under the covers have that mapped to the actual implementation.
     *             2. Use a defined, standard set of "extension point names" and map any extension to an
     *                actual function name under the covers.
     *
     * An example: In Enabler, all of the "public" functions are defined in the com.ibm.mashups.enabler.*
     * package. Since JavaScript does not have the concept of interfaces, the "public" objects contain no
     * logic whatsoever; empty properties and methods are defined to make up the abstract "interface". The
     * implementation is contained in a separate package, com.ibm.mm.enabler.*. We want extenders to use the
     * "public" API package and object name to declare extensions since it's part of the programming model.
     * To allow the public function names to transparently "map" to the private implementations, we provide
     * a mapping function to the extension registry that processes the Enabler extensions and maps any public
     * function referenced to it's "private" implementation.
     *
     * Another example: If you wanted to define a specific extension point for adding a new filter in a filter
     * chain, for example, using a mapping function could provide an easy way to do this. You could simply
     * define a property in the extension definition object called "additionalFilters" and provide a mapping function
     * to map "additionalFilters" to the appropriate function (say "addFilter").
     *
     * @constructor
     * @param {Function} mappingFn OPTIONAL the mapping function for mapping extension point names to actual function objects
     *
     */
    this.ExtensionRegistry = function(mappingFn) {
        //SETUP
        
        //Global object reference for use throughout the constructor.
        var global = window;
        
        //Logger for use throughout the constructor. Using this object abstracts
        //the logging process to allow this object to be as portable as possible. 
        var logger = {
            root: document.body,
            enabled: false,
            _queue: [],
            debug: function( /*Object*/obj) {
                if (this.enabled) {
                    //console.debug( obj );
                }
            },
            log: function( /*String*/msg) {
                /*var div = document.createElement( "DIV" );
                 div.innerHTML = msg;
                 
                 if ( this.root != null ) {
                 this._flushQueueIfNecessary();
                 this.root.appendChild( div );
                 } else {
                 this._queue.push( div );
                 }*/
                if (this.enabled) {
                    //console.log(msg);
                }
                //alert( msg );
            },
            _flushQueueIfNecessary: function() {
                var flushing = (this._queue.length > 0);
                while (this._queue.length > 0) {
                    this.root.appendChild(this._queue.shift());
                }
                
                if (flushing) {
                    this.log("----------------- END STMTS EXECUTED BEFORE ONLOAD ---------------");
                }
            }
        };
        
        var defaultMapping = function(name) {
            var names = name.split(".");
            var scope = null;
            var currentObj = window;
            var methodName = null;
            for (var i = 0; (i < names.length && currentObj); i++) {
                scope = currentObj;
                methodName = names[i];
                currentObj = scope[names[i]];
            }
            
            return [{
                object: scope,
                method: methodName
            }];
        };
        
        //If no mapping function is provided, we still need to provide a scope and
        //a method name from the given extension point name.
        if (!mappingFn) {
            mappingFn = defaultMapping;
        }
        
        //INTERNAL FUNCTIONS    
        
        //For...in loops don't work quite the way we want them to in JavaScript. We only want to loop over properties
        //that have been set by the user. Calling hasOwnProperty for each property on the object ensures that we don't try to process
        //a standard Object property (i.e. toString).
        function forInLoop( /*Object*/obj, /*Function*/ work) {
            for (var prop in obj) {
                if (Object.prototype.hasOwnProperty.call(obj,prop)) {
                    work(prop, obj[prop]);
                }
            }
        }
        
        //Determines if the specified object is an array. 
        function isArray(obj) {
            return (obj && obj.length && typeof(obj.length) == "number");
        }
        
        //Handles the case where a piece of work needs to be applied to an object but we allow the object to be passed 
        //singularly and in an array (for applying that work to a number of those objects).
        function applyToObjectOrArray( /*Object or Array*/obj, /*Function*/ work) {
            if (isArray(obj)) {
                for (var i = 0; i < obj.length; i++) {
                    work(obj[i]);
                }
            }
            else {
                work(obj);
            }
        }
        
        //INTERNAL OBJECTS
        
        //Maintains information about a particular function that has extensions attached. In particular, 
        //an instance of this object will keep track of the scope (or object the method belongs to), the
        //method name (in string format) and the value of the method at construction time. This is mainly
        //a convenience object because it gets cumbersome constantly typing:
        //         md.object[md.method] = ...
        //It also determines if we should be working with the object or the object's prototype. Finally,
        //this object provides an easy method to revert the value represented by the object and method to
        //be reverted to it's original value.
        function MethodDescriptor( /*Object*/obj, /*String*/ methodName) {
            this.method = methodName;
            this.object = (obj.prototype && obj.prototype[this.method]) ? obj.prototype : obj;
            this.originalValue = this.object[this.method];
            this.id = (this.object && this.object.declaredClass) ? this.object.declaredClass : this.originalValue.toString();
        }
        //Gets the current value of the method pointed to by this MethodDescriptor instance. Equivalent
        //to retrieving the value using bracket notation: this.object[ this.method ].
        MethodDescriptor.prototype.getMethod = function() {
            return this.object[this.method];
        };
        //Sets the current value of the method pointed to by this MethodDescriptor instance. The old value
        //is overwritten.
        MethodDescriptor.prototype.setMethod = function(value) {
            this.object[this.method] = value;
        };
        //Reverts the value of the method pointed to by this MethodDescriptor instance to it's original value.
        MethodDescriptor.prototype.revert = function() {
            this.setMethod(this.originalValue);
        };
        //Prints a JSON-like String representation of this MethodDescriptor instance.
        MethodDescriptor.prototype.toString = function() {
            return "{ object: " + this.object + ", method: " + this.method + ", originalValue: " + (this.originalValue ? this.originalValue.toString() : this.originalValue) + ", currentValue: " + this.getMethod().toString() + " }";
        };
        
        //Manages and controls all extensions attached to a given extension point. Keeps track of the extension
        //point name, function descriptor (MethodDescriptor), and all attached extensions. The before and after
        //extensions are stored in arrays because multiple handlers are allowed for each of those types of advice.
        //The around extension is stored as a variable because only one type of around advice is allowed at any
        //given point in time.  
        function ExtensionPointController( /*String*/extPtName, /*MethodDescriptor*/ originalFnDescr) {
            this.name = extPtName;
            this.fnDescriptor = originalFnDescr;
            this.extensions = {
                before: [],
                after: []
            };
        }
        
        //Adds a new layer to the extended function with the new piece of around advice. 
        ExtensionPointController.prototype.addAroundAdvice = function(handler) {
            var fn = this.fnDescriptor.getMethod();
            this.fnDescriptor.setMethod(function() {
                if (logger.enabled) {
                    logger.log("around advice entry ==> [ " + arguments + " ]");
                    logger.log("handler to execute: " + handler.toString());
                    logger.debug(handler);
                    logger.log("next function in the chain (executor parameter): " + fn.toString());
                    logger.debug(fn);
                }
                
                var me = this;
                var retVal = handler.call(this, function(args) {
                    return fn.apply(me, args);
                }, arguments);
                
                if (logger.enabled) {
                    logger.log("around advice exit ==> " + retVal);
                }
                return retVal;
            });
        };
        
        //Executes the appropriate extensions and the original function, if necessary. The executionScope is the
        //scope in which all advice (and the original function) is executed. fnArgs is an array of arguments that
        //are passed to each function executed in the given executionScope.
        ExtensionPointController.prototype.execute = function(executionScope, fnArgs) {
            if (logger.enabled) {
                logger.log("ExtensionPointController.execute entry ==> [ " + executionScope + ", " + fnArgs + " ] ");
                logger.debug(executionScope);
                logger.debug(fnArgs);
            }
            
            //Apparently, the arguments object is not a true array. This function copies the elements in the arguments
            //object into a true array so we can have access to the standard array operations.
            function argumentsToArray( /*Object{arguments}*/args) {
                var arr = [];
                for (var i = 0; i < args.length; i++) {
                    arr.push(args[i]);
                }
                return arr;
            }
            
            var retVal = null;
            
            if (logger.enabled) {
                logger.log("Executing extensions for \"" + this.name + "\"...");
            }
            
            var extensions = this.extensions;
            var descriptor = this.fnDescriptor;
            function doExecution( /*Array*/args) {
                if (logger.enabled) {
                    logger.log("doExecution entry ==> " + args);
                    logger.debug(args);
                }
                
                //Make sure we have a true array.
                if (args) {
                    args = argumentsToArray(args);
                }
                
                var returnValue = null;
                //Revert to the original value before executing the function. The original function might 
                //reference properties by name that will not be accessible if the original function is
                //executed out of the original context. Store the new value so it can be replaced later.
                var temp = descriptor.getMethod();
                descriptor.revert();
                
                //Execute the before advice.
                var beforeAdvice = extensions.before;
                if (logger.enabled) {
                    logger.log("Executing " + beforeAdvice.length + " before advice extensions.");
                }
                for (var i = 0; i < beforeAdvice.length; i++) {
                    if (logger.enabled) {
                        logger.log("beforeAdvice[" + i + "] = " + beforeAdvice[i].toString());
                        logger.debug(beforeAdvice[i]);
                    }
                    args = beforeAdvice[i].call(this, args);
                }
                
                //Execute the original function and store the return value.
                if (logger.enabled) {
                    logger.log("Executing the original function: " + descriptor.getMethod().toString());
                }
                returnValue = descriptor.getMethod().apply(this, args);
                
                //Execute the after advice
                var afterAdvice = extensions.after;
                if (logger.enabled) {
                    logger.log("Executing " + afterAdvice.length + " after advice extensions.");
                }
                //First parameter to after advice should be the return value.
                if (args) {
                    for (var j = 0; j < afterAdvice.length; j++) {
                        if (logger.enabled) {
                            logger.log("afterAdvice[" + j + "] = " + afterAdvice[j].toString());
                            logger.debug(afterAdvice[j]);
                            logger.log("Return value: " + returnValue);
                            logger.debug(returnValue);
                        }
                        //Make sure the return value is always up to date.
                        returnValue = afterAdvice[j].call(this, returnValue, args);
                    }
                }
                
                //Re-set the extended method.
                descriptor.setMethod(temp);
                
                if (logger.enabled) {
                    logger.log("doExecution exit ==> " + returnValue);
                }
                return returnValue;
            }
            
            retVal = doExecution.apply(executionScope, [fnArgs]);
            
            if (logger.enabled) {
                logger.log("ExtensionPointController.execute exit ==> " + retVal);
            }
            
            return retVal;
        };
        
        //PRIVATE VARIABLES AND FUNCTIONS
        
        //Store all controllers internally by extension point name.
        if (!ExtensionRegistry.prototype.extensionPointControllers) {
            ExtensionRegistry.prototype.extensionPointControllers = {};
        }
        //Retrieves the extension point controller instance for a particular extension point name. If a controller
        //doesn't yet exist, one will be created and initialized.
        function getExtensionPointController(extPtName, methodDescriptor) {
            if (logger.enabled) {
                logger.log("ExtensionRegistry.getExtensionPointController entry ==> [ " + extPtName + ", " + methodDescriptor + " ]");
            }
            
            var existingExtPt = ExtensionRegistry.prototype.extensionPointControllers[extPtName];
            if (!existingExtPt) {
                existingExtPt = {};
                ExtensionRegistry.prototype.extensionPointControllers[extPtName] = existingExtPt;
            }
            
            var existingCtrl = existingExtPt[methodDescriptor.id];
            
            if (!existingCtrl) {
                existingCtrl = new ExtensionPointController(extPtName, methodDescriptor);
                ExtensionRegistry.prototype.extensionPointControllers[extPtName][methodDescriptor.id] = existingCtrl;
                methodDescriptor.setMethod(function() {
                    if (logger.enabled) {
                        logger.log(extPtName + " entry ==> " + arguments);
                    }
                    var retVal = existingCtrl.execute(this, arguments);
                    if (logger.enabled) {
                        logger.log(extPtName + " exit ==> " + retVal);
                    }
                    return retVal;
                });
            }
            
            if (logger.enabled) {
                logger.log("ExtensionRegistry.getExtensionPointController exit ==> " + existingCtrl);
            }
            return existingCtrl;
        }
        //Processes a single extension definition. Expects the extension point name and the extension's type
        //and handler function. The mapping function will be applied to the extension point name before the
        //extension definition is processed.
        function processExtensionDefinition(name, value) {
            if (logger.enabled) {
                logger.log("processExtensionDefinition entry ==> [ " + name + ", " + value + " ]");
                logger.log("Extension: " + name + "; extension: { type: \"" + value.type + "\", handler: " + value.handler + " }");
            }
            var impls = mappingFn(name);
            impls = impls || null;
            if (impls === null) {
                impls = defaultMapping(name);
            }
            if (logger.enabled) {
                logger.log("impls=" + impls);
            }
            for (var i = 0; i < impls.length; i++) {
                var impl = impls[i];
                var md = new MethodDescriptor(impl.object, impl.method);
                
                if (md.originalValue) {
                    var extPtCtrl = getExtensionPointController(name, md);
                    if (logger.enabled) {
                        logger.log("Extension Point (after mapping function): " + md);
                    }
                    
                    switch (value.type) {
                        case "before":
                            if (logger.enabled) {
                                logger.log("wrapping in a before extension");
                            }
                            extPtCtrl.extensions.before.push(value.handler);
                            break;
                        case "after":
                            if (logger.enabled) {
                                logger.log("wrapping in an after extension");
                            }
                            extPtCtrl.extensions.after.push(value.handler);
                            break;
                        case "around":
                            if (logger.enabled) {
                                logger.log("wrapping in an around extension");
                            }
                            extPtCtrl.addAroundAdvice(value.handler);
                            break;
                        default:
                            if (logger.enabled) {
                                logger.log("Unrecognized extension type: " + value.type);
                            }
                    }
                }
                else {
                    if (console && console.log) {
                        console.log("WARNING! \"" + name + "\" does not exist! The provided extension will be ignored!");
                    }
                }
            }
            
            if (logger.enabled) {
                logger.log("processExtensionDefinition exit");
            }
        }
        
        function removeExtensionDefinition(name) {
            if (logger.enabled) {
                logger.log("removeExtensionDefinition entry ==> [ " + name + " ]");
            }
            var defs = ExtensionRegistry.prototype.extensionPointControllers[name];
            forInLoop(defs, function(name, value) {
                value.fnDescriptor.revert();
            });
            delete ExtensionRegistry.prototype.extensionPointControllers[name];
            
            if (logger.enabled) {
                logger.log("removeExtensionDefinition exit");
            }
        }
        
        //PUBLIC METHODS
        
        /**
         * Processes an extension definition or set of extension definitions. An extension definition object
         * takes the following form:
         *
         * {
         *     "FunctionName" : {
         *                        type: "after" | "around" | "before",
         *                        handler: function () { ...extension function... }
         *                      },
         *  "FunctionName2" : [
         *                      {
         *                          type: "after" | "around" | "before",
         *                          handler: function () { ...extension function... }
         *                      },
         *                      {
         *                          type: "after" | "around" | "before",
         *                          handler: function () { ...extension function... }
         *                      }
         *                    ],
         * ...
         * }
         *
         * The type property refers to the type of "advice" the handler property should be executed
         * as.
         *
         *     "after"    : the handler is executed AFTER the extended function
         *  "around"   : the handler is executed INSTEAD of the extended function but has the option
         *               to execute the extended function
         *  "before"   : the handler is executed BEFORE the extended function
         *
         * Multiple extensions for a given extension point can be defined either by making the value
         * in a single definition object into an array (see FunctionName2 above) OR by passing an array
         * of definition objects into this function. The FunctionName property must be unique within a
         * definition object.
         *
         * BEFORE and AFTER advice can have multiple handlers attached because neither affects whether or
         * not the original function is executed or not. AROUND advice can only have one handler at a given
         * time. The last AROUND advice handler set for a particular extension point is the handler that
         * will be executed when the around advice is triggered for that extension point.
         *
         * @param {Object or Array} extDef a single definition object OR an array of definition objects
         */
        this.process = function(extDef) {
            if (logger.enabled) {
                logger.log("ExtensionRegistry.process entry ==>" + extDef);
            }
            
            applyToObjectOrArray(extDef, function(extDef) {
                forInLoop(extDef, function(name, value) {
                    applyToObjectOrArray(value, function(obj) {
                        processExtensionDefinition(name, obj);
                    });
                });
            });
            
            if (logger.enabled) {
                logger.log("ExtensionRegistry.process exit");
            }
        };
        /**
         * Indicates whether the mapping should continue or stop.
         */
        this.disconnect = function(extName) {
            if (logger.enabled) {
                logger.log("ExtensionRegistry.disconnect entry ==>" + extName);
            }
            
            removeExtensionDefinition(extName);
            
            if (logger.enabled) {
                logger.log("ExtensionRegistry.disconnect exit");
            }
        };
        
        this.enableTracking = function(extensionConfig) {
            var implTracker = new ImplementationTracker(true);
            var DojoExtensions = {
                "dojo.declare": {
                    type: "before",
                    handler: function ( fnArgs ) {
                        implTracker.add( fnArgs[0], fnArgs[1] );
                        return fnArgs;
                    }   
                }
            };
            this.process( DojoExtensions );
            var mappingFn = function ( name ) {
                return implTracker.getImpls( name );
            };
            
            dojo.addOnLoad( function() {
                if ( extensionConfig && extensionConfig.extensions && extensionConfig.extensions.length > 0 ) {
                    var configVar = {};
                    function forInLoop ( /*Object*/obj, /*Function*/work ) {
                        for ( var prop in obj ) {
                            if ( obj.hasOwnProperty( prop ) ) {
                                work( prop, obj[prop] );
                            }   
                        }   
                    }
                    for ( var i = 0; i < extensionConfig.extensions.length; i++ ) {
                        forInLoop( extensionConfig.extensions[i], function ( name, value ) {
                            configVar[name] = value;
                        });
                    }
                    var extReg = new ExtensionRegistry( mappingFn );
                    extReg.process( extensionConfig.extensions );
                }
            });
        };

        this.disableTracking = function() {
             this.disconnect( "dojo.declare" );
        };
    };
}

if ( typeof ImplementationTracker == "undefined" ) {
    // summary: constructor for the object that tracks inheritance using the dojo conventions
    // deep: OPTIONAL if true, then the full inheritance hierarchy (parent-child-grandchild-etc...) is tracked. if false, only 
    //        direct (parent-child) inheritance is tracked. Note that there may be a performance penalty with tracking the full 
    //        hierarchy depending on the number of objects declared and the complexity of the inheritance tree. 
    this.ImplementationTracker = function ( /*boolean?*/deep ) {
        //SETUP
        var logger = {
            enabled: false,
            log: function ( msg ) {
                if (this.enabled) {
                    //console.log(msg);
                }
            },
            debug: function ( obj ) {
                if (this.enabled) {
                    //console.debug(obj);
                }
            }
        };
        
        //INTERNAL FUNCTIONS
        
        //For...in loops don't work quite the way we want them to in JavaScript. We only want to loop over properties
        //that have been set by the user. Calling hasOwnProperty for each property on the object ensures that we don't try to process
        //a standard Object property (i.e. toString).
        function forInLoop ( /*Object*/obj, /*Function*/work ) {
            for (var prop in obj ) {
                if (Object.prototype.hasOwnProperty.call( obj, prop ) ) {
                    work( prop, obj[prop] );
                }    
            }    
        }
        
        //PRIVATE FUNCTIONS AND PROPERTIES
        
        //Determines, from a given extension point name, the object the method belongs to and the
        //name of the method. The return value takes the form: { object: [Object], method: String }.
        function getMethodImpl( /*String*/name ) {
            var names = name.split(".");
            var scope = null;
            var currentObj = window;
            var methodName = null;
            for ( var i = 0, l = names.length; (i < l && currentObj); i++ ) {
                scope = currentObj;
                methodName = names[i];
                currentObj = scope[ names[i] ];
            }
            
            return { object: scope, method: methodName };
        }
        
        //Internal store for the objects we are tracking implementations of. The structure of the object is a flat object with a property for
        //each class/interface defined. Each property's value contains an instance of ObjectInfo which tracks the descendants (all if deep, direct only if not deep)
        //and the parents of an object.
        var objects = {};
        
        //Convenience object for tracking any implementors.
        function ObjectInfo ( /*String*/name ) {
            this.className = name;
            this.parents = [];
            this.implementors = [];
        }
        
        //PUBLIC FUNCTIONS AND PROPERTIES
        
        /**
         * Dumps the current object store contents to the specified log location.
         */
        ImplementationTracker.prototype.dumpObjects = function () {
            if (logger.enabled) {
                forInLoop(objects, function(name, value){
                    var logStr = "\"" + name + "\" : [ ";
                    
                    for (var i = 0, l = value.implementors.length; i < l; i++) {
                        logStr += value.implementors[i].className;
                        if (i != (l - 1)) {
                            logStr += ", ";
                        }
                    }
                    
                    logStr += " ];";
                    
                    logger.log(logStr);
                });
            }
        };
        
        /**
         * Adds the implementor to each of the parent objects all the way up the inheritance tree.
         */
        function processAncestors( /*Array*/parents, /*ObjectInfo*/ implementor ) {
            if ( parents ) {
                var len = parents.length, parent;
                for ( var i = 0; i < len; i++ ) {
                    parent = objects[ parents[i] ];
                    if ( parent ) {
                        parent.implementors.push( implementor );
                    }
                    processAncestors( parents[i].parents, implementor );
                }
            }
        }
        
        /**
         * Adds an object to the object store.
         * 
         * @param {String} objName the fully qualified name of the object
         * @param {Array} parents the array of objects that the current object inherits from
         */
        ImplementationTracker.prototype.add = function ( /*String*/objName, /*Array{String}*/parents ) {
            if (logger.enabled) {
                logger.log("ImplementationTracker.add entry ==> " + [objName, parents]);
            }
            if ( !objects[objName] ) {
                if ( parents && typeof parents == "function" ) {
                    parents = [ parents ];
                }

                var info = new ObjectInfo( objName );
                objects[ objName ] = info;
                for ( var i = 0, l = parents ? parents.length : 0; i < l; i++ ) {
                    if ( parents[i].prototype ) {
                        var parentClassName = parents[i].prototype.declaredClass;
                        info.parents.push( parentClassName );
                        var parentObjectInfo = objects[ parentClassName ];
                        
                        if (logger.enabled) {
                            logger.log("Processing parent classname: " + parentClassName);
                            logger.log("parent exists in object store? " + parentObjectInfo);
                        }
                        if ( parentObjectInfo ) {
                            parentObjectInfo.implementors.push( info );
                            if ( deep ) {
                                processAncestors( parentObjectInfo.parents, info );
                            }
                        }
                        else {
                            parentObjectInfo = new ObjectInfo( parentClassName );
                            parentObjectInfo.implementors.push( info );
                        }
                    }    
                }
            }
            if (logger.enabled) {
                logger.log("ImplementationTracker.add exit");
            }
        };
        
        /**
         * Retrieves the implementations associated with the given object name.
         *  
         * @param {String} publicName the public name of the object
         * @return {Array{ object: Object, method: String }} an array of implementations associated with the given object name.
         */
        ImplementationTracker.prototype.getImpls = function ( /*String*/publicName ) {
            if ( logger.enabled ) { logger.log( "ImplementationTracker.getImpls entry ==> " + publicName ); }
            var className = publicName.substring( 0, publicName.lastIndexOf( "." ) );
            var methodName = publicName.substring( publicName.lastIndexOf( "." ) + 1  );
            if ( logger.enabled ) { logger.log( "className = " + className + "; methodName = " + methodName ); }
            //Make sure implementors exist AND that the public function exists!
            var methodImpl = getMethodImpl( publicName );
            methodImpl = methodImpl || null;
            if ( !objects[ className ] || methodImpl === null ) {
                if ( logger.enabled ) { logger.log( "ImplementationTracker.getImpls exit ==> null " ); }
                return null;
            }
            var impls = objects[ className ].implementors;

            if ( logger.enabled ) { logger.log( "impls = " + impls ); }

            var retVal = [];
            for ( var i = 0, l = impls.length; i < l; i++ ) {
                if (logger.enabled) {
                    logger.log("impls[i]=" + impls[i].className);
                }
                methodImpl = getMethodImpl( impls[i].className + "." + methodName );
                methodImpl = methodImpl || null;
                if ( methodImpl !== null ) {
                    retVal.push( methodImpl );
                }
            }

            if ( logger.enabled ) { logger.log( "ImplementationTracker.getImpls exit ==> " + retVal ); }
            return retVal; // Array of { object: Object, method: String }
        };
    };
}

